# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""OpenFL Assigner Module."""

from openfl.component.assigner.assigner import Assigner
from openfl.component.assigner.random_grouped_assigner import RandomGroupedAssigner
from openfl.component.assigner.static_grouped_assigner import StaticGroupedAssigner



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Random grouped assigner module."""

import numpy as np

from openfl.component.assigner import Assigner


class RandomGroupedAssigner(Assigner):
    r"""The task assigner maintains a list of tasks.

    Also it decides the policy for which collaborator should run those tasks
    There may be many types of policies implemented, but a natural place to
    start is with a:

        - RandomGroupedAssigner :
            Given a set of task groups, and a percentage,
            assign that task group to that percentage of collaborators in the
            federation.
            After assigning the tasks to collaborator, those tasks should be
            carried out each round (no reassignment between rounds).
        - GroupedAssigner :
            Given task groups and a list of collaborators that belong to that
            task group, carry out tasks for each round of experiment.

    Attributes:
        task_groups* (list of object): Task groups to assign.

    .. note::
        \* - Plan setting.
    """

    with_selected_task_group = Assigner.with_selected_task_group

    def __init__(self, task_groups, **kwargs):
        """Initializes the RandomGroupedAssigner.

        Args:
            task_groups (list of object): Task groups to assign.
            **kwargs: Additional keyword arguments, including mode.
        """
        self.task_groups = task_groups
        super().__init__(**kwargs)

    @with_selected_task_group
    def define_task_assignments(self):
        """Define task assignments for each round and collaborator.

        This method uses the assigner function to assign tasks to
        collaborators for each round.
        It also maps tasks to their respective aggregation functions.

        Args:
            None

        Returns:
            None
        """
        assert np.abs(1.0 - np.sum([group["percentage"] for group in self.task_groups])) < 0.01, (
            "Task group percentages must sum to 100%"
        )

        # Start by finding all of the tasks in all specified groups
        self.all_tasks_in_groups = list(
            {task for group in self.task_groups for task in group["tasks"]}
        )

        # Initialize the map of collaborators for a given task on a given round
        for task in self.all_tasks_in_groups:
            self.collaborators_for_task[task] = {i: [] for i in range(self.rounds)}

        for col in self.authorized_cols:
            self.collaborator_tasks[col] = {i: [] for i in range(self.rounds)}

        col_list_size = len(self.authorized_cols)
        for round_num in range(self.rounds):
            randomized_col_idx = np.random.choice(
                len(self.authorized_cols),
                len(self.authorized_cols),
                replace=False,
            )
            col_idx = 0
            for group in self.task_groups:
                num_col_in_group = int(group["percentage"] * col_list_size)
                rand_col_group_list = [
                    self.authorized_cols[i]
                    for i in randomized_col_idx[col_idx : col_idx + num_col_in_group]
                ]
                self.task_group_collaborators[group["name"]] = rand_col_group_list
                for col in rand_col_group_list:
                    self.collaborator_tasks[col][round_num] = group["tasks"]
                # Now populate reverse lookup of tasks->group
                for task in group["tasks"]:
                    # This should append the list of collaborators performing
                    # that task
                    self.collaborators_for_task[task][round_num] += rand_col_group_list
                col_idx += num_col_in_group
            assert col_idx == col_list_size, "Task groups were not divided properly"

    def get_tasks_for_collaborator(self, collaborator_name, round_number):
        """Get tasks for a specific collaborator in a specific round.

        Args:
            collaborator_name (str): Name of the collaborator.
            round_number (int): Round number.

        Returns:
            list: List of tasks for the collaborator in the specified round.
        """
        return self.collaborator_tasks[collaborator_name][round_number]

    def get_collaborators_for_task(self, task_name, round_number):
        """Get collaborators for a specific task in a specific round.

        Args:
            task_name (str): Name of the task.
            round_number (int): Round number.

        Returns:
            list: List of collaborators for the task in the specified round.
        """
        return self.collaborators_for_task[task_name][round_number]



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
from openfl.callbacks.callback import Callback
from openfl.callbacks.memory_profiler import MemoryProfiler
from openfl.callbacks.metric_writer import MetricWriter


class CallbackList(Callback):
    """An ensemble of callbacks.

    This class allows multiple callbacks to be used together, by sequentially
    calling each callback's respective methods.

    Attributes:
        callbacks: A list of `openfl.callbacks.Callback` instances.
        add_memory_profiler: If True, adds a `MemoryProfiler` callback to the list.
        add_metric_writer: If True, adds a `MetricWriter` callback to the list.
        tensor_db: Optional `TensorDB` instance of the respective participant.
            If provided, callbacks can access TensorDB for various actions.
        params: Additional parameters saved for use within the callbacks.
    """

    def __init__(
        self,
        callbacks: list,
        add_memory_profiler=False,
        add_metric_writer=False,
        tensor_db=None,
        **params,
    ):
        super().__init__()
        self.callbacks = list(_flatten(callbacks)) if callbacks else []

        self._add_default_callbacks(add_memory_profiler, add_metric_writer)

        self.set_tensor_db(tensor_db)
        self.set_params(params)

    def set_params(self, params):
        self.params = params
        if params:
            for callback in self.callbacks:
                callback.set_params(params)

    def set_tensor_db(self, tensor_db):
        self.tensor_db = tensor_db
        if tensor_db:
            for callback in self.callbacks:
                callback.set_tensor_db(tensor_db)

    def _add_default_callbacks(self, add_memory_profiler, add_metric_writer):
        """Add default callbacks to callbacks list if not already present."""
        self._memory_profiler = None
        self._metric_writer = None

        for cb in self.callbacks:
            if isinstance(cb, MemoryProfiler):
                self._memory_profiler = cb
            if isinstance(cb, MetricWriter):
                self._metric_writer = cb

        if add_memory_profiler and self._memory_profiler is None:
            self._memory_profiler = MemoryProfiler()
            self.callbacks.append(self._memory_profiler)

        if add_metric_writer and self._metric_writer is None:
            self._metric_writer = MetricWriter()
            self.callbacks.append(self._metric_writer)

    def on_round_begin(self, round_num: int, logs=None):
        for callback in self.callbacks:
            callback.on_round_begin(round_num, logs)

    def on_round_end(self, round_num: int, logs=None):
        for callback in self.callbacks:
            callback.on_round_end(round_num, logs)

    def on_experiment_begin(self, logs=None):
        for callback in self.callbacks:
            callback.on_experiment_begin(logs=logs)

    def on_experiment_end(self, logs=None):
        for callback in self.callbacks:
            callback.on_experiment_end(logs)

    def on_task_begin(self, task_name: str, round_num: int, logs=None):
        for callback in self.callbacks:
            callback.on_task_begin(task_name, round_num, logs)

    def on_task_end(self, task_name: str, round_num: int, logs=None):
        for callback in self.callbacks:
            callback.on_task_end(task_name, round_num, logs)


def _flatten(l):
    """Flatten a possibly-nested tree of lists."""
    if not isinstance(l, (list, tuple)):
        return [l]
    for elem in l:
        if isinstance(elem, list):
            yield from _flatten(elem)
        else:
            yield elem



# Copyright (C) 2020-2024 Intel Corporation
# Licensed subject to the terms of the separately executed evaluation license agreement between Intel Corporation and you.

import numpy as np

from typing import Iterator, Tuple

from openfl.federated import PyTorchTaskRunner
from openfl.utilities import Metric

import torch.optim as optim
import torch.nn.functional as F
from src.cnn_model import DigitRecognizerCNN, train_epoch, validate

class TemplateTaskRunner(PyTorchTaskRunner):
    """Template Task Runner for PyTorch.

    This class should be used as a template to create a custom Task Runner for your specific model and training workload.
    After generating this template, you should:
    1. Define your model, optimizer, and loss function as you would in PyTorch. PyTorchTaskRunner inherits from torch.nn.Module.
    2. Implement the `train_` and `validate_` functions to define a single train and validate epoch of your workload.
    3. Modify the `plan.yaml` file to use this Task Runner.

    The `plan.yaml` modifications should be done under the `<workspace>/plan/plan.yaml` section:
    ```
    task_runner:
        defaults : plan/defaults/task_runner.yaml
        template: src.taskrunner.TemplateTaskRunner # Modify this line appropriately if you change the class name
        settings:
            # Add additional arguments that you wish to pass through `__init__`
    ```

    Define the `forward` method of this class as a forward pass through the model.
    """

    def __init__(self, device="cpu", **kwargs):
        """Initialize the Task Runner.

        Args:
            device: The hardware device to use for training (Default = "cpu").
            **kwargs: Additional arguments that may be defined in `plan.yaml`
        """
        super().__init__(device=device, **kwargs)

        # Define the model
        self.model = DigitRecognizerCNN()
        self.to(device)

        # Define the optimizer
        self.optimizer = optim.Adam(self.model.parameters(), lr=1e-4)

        # Define the loss function
        self.loss_fn = F.cross_entropy

    def forward(self, x):
        """Forward pass of the model.

        Args:
            x: Data input to the model for the forward pass.

        Returns:
            The output of the model's forward pass.
        """
        return self.model(x)

    def train_(
        self, train_dataloader: Iterator[Tuple[np.ndarray, np.ndarray]]
    ) -> Metric:
        """Single Training epoch.

        Args:
            train_dataloader: Train dataset batch generator. Yields (samples, targets) tuples
                              of size = `self.train_dataloader.batch_size`.

        Returns:
            Metric: An object containing the name of the metric and its value as an np.ndarray.
        """
        # Implement training logic here and return a Metric object with the training loss.
        # Replace the following placeholder with actual training code.

        loss = train_epoch(self.model, self.optimizer, self.loss_fn, train_dataloader, self.device)
        return Metric(name="crossentropy_loss", value=np.array(loss))

    def validate_(
        self, validation_dataloader: Iterator[Tuple[np.ndarray, np.ndarray]]
    ) -> Metric:
        """Single validation epoch.

        Args:
            validation_dataloader: Validation dataset batch generator. Yields (samples, targets) tuples.
                                   of size = `self.validation_dataloader.batch_size`.

        Returns:
            Metric: An object containing the name of the metric and its value as an np.ndarray.
        """
        # Implement validation logic here and return a Metric object with the validation accuracy.
        # Replace the following placeholder with actual validation code.

        accuracy = validate(self.model, validation_dataloader, self.device)
        return Metric(name="accuracy", value=np.array(accuracy))



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


max_metadata_size = 32 * 2**20
max_message_length = 2**30

channel_options = [
    ("grpc.max_metadata_size", max_metadata_size),
    ("grpc.max_send_message_length", max_message_length),
    ("grpc.max_receive_message_length", max_message_length),
]



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Straggler handling module."""

import threading
import time
from abc import ABC, abstractmethod
from logging import getLogger
from typing import Callable

import numpy as np

logger = getLogger(__name__)


class StragglerPolicy(ABC):
    """Federated Learning straggler handling interface."""

    @abstractmethod
    def start_policy(self, **kwargs) -> None:
        """
        Start straggler handling policy for collaborator for a particular round.
        NOTE: Refer CutoffTimePolicy class for reference.

        Args:
            **kwargs
        """
        raise NotImplementedError

    @abstractmethod
    def reset_policy_for_round(self) -> None:
        """Reset policy for the next round."""
        raise NotImplementedError

    @abstractmethod
    def straggler_cutoff_check(
        self, num_collaborators_done: int, num_all_collaborators: int, **kwargs
    ) -> bool:
        """
        Determines whether the round should end early when straggler policy conditions are met.

        Args:
            num_collaborators_done: int
                Number of collaborators finished.
            num_all_collaborators: int
                Total number of collaborators.

        Returns:
            bool: True if it is time to end the round early, False otherwise.

        Raises:
            NotImplementedError: This method must be implemented by a subclass.
        """
        raise NotImplementedError


class WaitForAllPolicy(StragglerPolicy):
    """
    This policy waits for all collaborators.
    """

    def start_policy(self, **kwargs) -> None:
        """
        Nothing needed.
        """
        pass

    def reset_policy_for_round(self) -> None:
        """
        Nothing needed.
        """
        pass

    def straggler_cutoff_check(
        self, num_collaborators_done: int, num_all_collaborators: int, **kwargs
    ) -> bool:
        """
        Checks if all collaborators have sent task results for the round.
        """
        return num_all_collaborators == num_collaborators_done


class CutoffTimePolicy(StragglerPolicy):
    """Cutoff time based Straggler Handling function."""

    def __init__(
        self, round_start_time=None, straggler_cutoff_time=np.inf, minimum_reporting=1, **kwargs
    ):
        """
         Initialize a CutoffTimePolicy object.

        Args:
            round_start_time (optional): The start time of the round. Defaults
                to None.
            straggler_cutoff_time (float, optional): The cutoff time for
                stragglers. Defaults to np.inf.
            minimum_reporting (int, optional): The minimum number of
                collaborators that should report before moving to the next round.
                Defaults to 1.
            **kwargs: Variable length argument list.
        """
        if minimum_reporting <= 0:
            raise ValueError("minimum_reporting must be >0")

        self.round_start_time = round_start_time
        self.straggler_cutoff_time = straggler_cutoff_time
        self.minimum_reporting = minimum_reporting
        self.is_timer_started = False

        if self.straggler_cutoff_time == np.inf:
            logger.warning(
                "CutoffTimePolicy is disabled as straggler_cutoff_time is set to np.inf."
            )

    def reset_policy_for_round(self) -> None:
        """Reset timer for the next round."""
        if hasattr(self, "timer"):
            self.timer.cancel()
        self.is_timer_started = False

    def start_policy(self, callback: Callable) -> None:
        """
        Start time-based straggler handling policy for collaborator for
        a particular round.

        Args:
            callback: Callable
                Callback function for when straggler_cutoff_time elapses
        """
        # If straggler_cutoff_time is set to infinity
        # or if the timer is already running,
        # do not start the policy.
        if self.straggler_cutoff_time == np.inf or self.is_timer_started:
            return

        self.round_start_time = time.time()
        self.timer = threading.Timer(
            self.straggler_cutoff_time,
            callback,
        )
        self.timer.daemon = True
        self.timer.start()
        self.is_timer_started = True

    def straggler_cutoff_check(
        self,
        num_collaborators_done: int,
        num_all_collaborators: int,
    ) -> bool:
        """
        If minimum_reporting collaborators have reported results within
        straggler_cutoff_time then return True, otherwise False.

        Args:
            num_collaborators_done: int
                Number of collaborators finished.
            num_all_collaborators: int
                Total number of collaborators.

        Returns:
            bool: True if the straggler cutoff conditions are met, False otherwise.
        """

        # if straggler time has not expired then
        # wait for ALL collaborators to report results.
        if not self.__straggler_time_expired():
            return num_all_collaborators == num_collaborators_done

        # Time has expired
        # Check if minimum_reporting collaborators have reported results
        elif self.__minimum_collaborators_reported(num_collaborators_done):
            logger.info(
                f"{num_collaborators_done} collaborators have reported results. "
                "Applying cutoff policy and proceeding with end of round."
            )
            return True
        else:
            logger.info(
                f"Waiting for minimum {self.minimum_reporting} collaborator(s) to report results."
            )
            return False

    def __straggler_time_expired(self) -> bool:
        """Check if the straggler time has expired.

        Returns:
            bool: True if the straggler time has expired, False otherwise.
        """
        return self.round_start_time is not None and (
            (time.time() - self.round_start_time) > self.straggler_cutoff_time
        )

    def __minimum_collaborators_reported(self, num_collaborators_done) -> bool:
        """Check if the minimum number of collaborators have reported.

        Args:
            num_collaborators_done (int): The number of collaborators that
                have reported.

        Returns:
            bool: True if the minimum number of collaborators have reported,
                False otherwise.
        """
        return num_collaborators_done >= self.minimum_reporting


class PercentagePolicy(StragglerPolicy):
    """Percentage based Straggler Handling function."""

    def __init__(self, percent_collaborators_needed=1.0, minimum_reporting=1, **kwargs):
        """Initialize a PercentagePolicy object.

        Args:
            percent_collaborators_needed (float, optional): The percentage of
                collaborators needed. Defaults to 1.0.
            minimum_reporting (int, optional): The minimum number of
                collaborators that should report. Defaults to 1.
            **kwargs: Variable length argument list.
        """
        if minimum_reporting <= 0:
            raise ValueError("minimum_reporting must be >0")

        self.percent_collaborators_needed = percent_collaborators_needed
        self.minimum_reporting = minimum_reporting

    def reset_policy_for_round(self) -> None:
        """Not required in PercentagePolicy."""
        pass

    def start_policy(self, **kwargs) -> None:
        """Not required in PercentagePolicy."""
        pass

    def straggler_cutoff_check(
        self,
        num_collaborators_done: int,
        num_all_collaborators: int,
    ) -> bool:
        """
        If percent_collaborators_needed and minimum_reporting collaborators have
        reported results, then it is time to end round early.

        Args:
            num_collaborators_done (int): The number of collaborators that
                have reported.
            all_collaborators (list): All the collaborators.

        Returns:
            bool: True if the straggler cutoff conditions are met, False
                otherwise.
        """
        return (
            num_collaborators_done >= self.percent_collaborators_needed * num_all_collaborators
        ) and self.__minimum_collaborators_reported(num_collaborators_done)

    def __minimum_collaborators_reported(self, num_collaborators_done) -> bool:
        """Check if the minimum number of collaborators have reported.

        Args:
            num_collaborators_done (int): The number of collaborators that
                have reported.

        Returns:
            bool: True if the minimum number of collaborators have reported,
                False otherwise.
        """
        return num_collaborators_done >= self.minimum_reporting



from openfl.protocols import base_pb2 as _base_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class MessageHeader(_message.Message):
    __slots__ = ("sender", "receiver", "federation_uuid", "single_col_cert_common_name")
    SENDER_FIELD_NUMBER: _ClassVar[int]
    RECEIVER_FIELD_NUMBER: _ClassVar[int]
    FEDERATION_UUID_FIELD_NUMBER: _ClassVar[int]
    SINGLE_COL_CERT_COMMON_NAME_FIELD_NUMBER: _ClassVar[int]
    sender: str
    receiver: str
    federation_uuid: str
    single_col_cert_common_name: str
    def __init__(self, sender: _Optional[str] = ..., receiver: _Optional[str] = ..., federation_uuid: _Optional[str] = ..., single_col_cert_common_name: _Optional[str] = ...) -> None: ...

class PingRequest(_message.Message):
    __slots__ = ("header",)
    HEADER_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ...) -> None: ...

class PingResponse(_message.Message):
    __slots__ = ("header",)
    HEADER_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ...) -> None: ...

class GetTasksRequest(_message.Message):
    __slots__ = ("header",)
    HEADER_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ...) -> None: ...

class Task(_message.Message):
    __slots__ = ("name", "function_name", "task_type", "apply_local")
    NAME_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_NAME_FIELD_NUMBER: _ClassVar[int]
    TASK_TYPE_FIELD_NUMBER: _ClassVar[int]
    APPLY_LOCAL_FIELD_NUMBER: _ClassVar[int]
    name: str
    function_name: str
    task_type: str
    apply_local: bool
    def __init__(self, name: _Optional[str] = ..., function_name: _Optional[str] = ..., task_type: _Optional[str] = ..., apply_local: bool = ...) -> None: ...

class GetTasksResponse(_message.Message):
    __slots__ = ("header", "round_number", "tasks", "sleep_time", "quit")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    ROUND_NUMBER_FIELD_NUMBER: _ClassVar[int]
    TASKS_FIELD_NUMBER: _ClassVar[int]
    SLEEP_TIME_FIELD_NUMBER: _ClassVar[int]
    QUIT_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    round_number: int
    tasks: _containers.RepeatedCompositeFieldContainer[Task]
    sleep_time: int
    quit: bool
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ..., round_number: _Optional[int] = ..., tasks: _Optional[_Iterable[_Union[Task, _Mapping]]] = ..., sleep_time: _Optional[int] = ..., quit: bool = ...) -> None: ...

class GetAggregatedTensorRequest(_message.Message):
    __slots__ = ("header", "tensor_name", "round_number", "report", "tags", "require_lossless")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    TENSOR_NAME_FIELD_NUMBER: _ClassVar[int]
    ROUND_NUMBER_FIELD_NUMBER: _ClassVar[int]
    REPORT_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    REQUIRE_LOSSLESS_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    tensor_name: str
    round_number: int
    report: bool
    tags: _containers.RepeatedScalarFieldContainer[str]
    require_lossless: bool
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ..., tensor_name: _Optional[str] = ..., round_number: _Optional[int] = ..., report: bool = ..., tags: _Optional[_Iterable[str]] = ..., require_lossless: bool = ...) -> None: ...

class GetAggregatedTensorResponse(_message.Message):
    __slots__ = ("header", "round_number", "tensor")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    ROUND_NUMBER_FIELD_NUMBER: _ClassVar[int]
    TENSOR_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    round_number: int
    tensor: _base_pb2.NamedTensor
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ..., round_number: _Optional[int] = ..., tensor: _Optional[_Union[_base_pb2.NamedTensor, _Mapping]] = ...) -> None: ...

class TaskResults(_message.Message):
    __slots__ = ("header", "round_number", "task_name", "data_size", "tensors")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    ROUND_NUMBER_FIELD_NUMBER: _ClassVar[int]
    TASK_NAME_FIELD_NUMBER: _ClassVar[int]
    DATA_SIZE_FIELD_NUMBER: _ClassVar[int]
    TENSORS_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    round_number: int
    task_name: str
    data_size: int
    tensors: _containers.RepeatedCompositeFieldContainer[_base_pb2.NamedTensor]
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ..., round_number: _Optional[int] = ..., task_name: _Optional[str] = ..., data_size: _Optional[int] = ..., tensors: _Optional[_Iterable[_Union[_base_pb2.NamedTensor, _Mapping]]] = ...) -> None: ...

class SendLocalTaskResultsResponse(_message.Message):
    __slots__ = ("header",)
    HEADER_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ...) -> None: ...

class InteropMessage(_message.Message):
    __slots__ = ("header", "message", "metadata")
    class MetadataEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    HEADER_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    header: MessageHeader
    message: _base_pb2.DataStream
    metadata: _containers.ScalarMap[str, str]
    def __init__(self, header: _Optional[_Union[MessageHeader, _Mapping]] = ..., message: _Optional[_Union[_base_pb2.DataStream, _Mapping]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...



# Copyright (C) 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Histogram module."""

import numpy as np

from openfl.interface.aggregation_functions.core import AggregationFunction


class AggregateHistogram(AggregationFunction):
    """Histogram aggregation."""

    def call(self, local_tensors, *_) -> np.ndarray:
        """
        Aggregates a list of local histograms into a single global histogram.

        This method takes a list of objects, each containing a `tensor` attribute
        (which represents a local histogram as a numpy array), and combines them
        into a single aggregated histogram. A histogram is a representation of
        the distribution of data, typically used to count occurrences of values
        within specified ranges (bins). Aggregating histograms involves summing
        up the corresponding bins across all local histograms to produce a global
        view of the data distribution.

            local_tensors (list): A list of objects, where each object has a
                `tensor` attribute that is a numpy array representing a local
                histogram.

            np.ndarray: The aggregated histogram as a numpy array. Each bin in
            the resulting histogram is the sum of the corresponding bins from
            all input histograms. If the input list is empty, raises a
            ValueError indicating that the result is empty.

        Raises:
            ValueError: If the input list `local_tensors` is empty, indicating
            that there are no histograms to aggregate.
        """

        if not local_tensors:
            raise ValueError("Histogram result is empty.")

        agg_histogram = np.zeros_like(local_tensors[0].tensor)
        for local_tensor in local_tensors:
            agg_histogram += local_tensor.tensor
        return agg_histogram



# Copyright (C) 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""


from src.runner_fa import FederatedAnalyticsTaskRunner
import numpy as np

class IrisHistogram(FederatedAnalyticsTaskRunner):
    """
    Taskrunner class used to perform federated analytics on the Iris dataset by generating histograms for specified columns.
    Methods
    -------
    __init__(**kwargs)
        Initializes the IrisHistogram instance with the provided keyword arguments.
    analytics(columns, **kwargs)
        Performs analytics on the specified columns and returns a dictionary of histograms.
    compute_hist(df, col_name)
        Computes the histogram for a specified column in the dataframe.
    """


    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def analytics_task(self, columns, **kwargs):
        """
        Perform analytics on the specified columns and compute histograms.
        Args:
            columns (list): List of column names to analyze.
            **kwargs: Additional keyword arguments.
        Returns:
            dict: A dictionary where keys are column names and values are histograms.
        """
        # query data
        data = self.data_loader.query(columns)
        histograms = {}
        for column in columns:
            hist, bins = self.compute_hist(data, column)
            histograms[column + " histogram"] = hist
            histograms[column + " bins"] = bins
        return histograms

    def compute_hist(self, data, col_name):
        """
        Compute the histogram of a specified column in a DataFrame.
        Args:
            data (pandas.DataFrame): The DataFrame containing the data.
            col_name (str): The name of the column for which to compute the histogram.
        Returns:
            tuple: A tuple containing the histogram and bin edges as numpy arrays.
        """
        hist, bins = np.histogram(data[col_name], bins=np.linspace(2.0, 10.0, 10))
        return hist, bins



# Copyright (C) 2020-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import os.path
import pickle
from typing import Any, Callable, Optional, Tuple

import numpy as np
from PIL import Image

from torchvision.datasets.utils import check_integrity, download_and_extract_archive
from torchvision.datasets.vision import VisionDataset


class CIFAR10(VisionDataset):
    """`CIFAR10 <https://www.cs.toronto.edu/~kriz/cifar.html>`_ Dataset.

    Args:
        root (string): Root directory of dataset where directory
            ``cifar-10-batches-py`` exists or will be saved to if download is set to True.
        train (bool, optional): If True, creates dataset from training set, otherwise
            creates from test set.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.

    """

    base_folder = "cifar-10-batches-py"
    url = "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
    filename = "cifar-10-python.tar.gz"
    tgz_md5 = "c58f30108f718f92721af3b95e74349a"
    train_list = [
        ["data_batch_1", "c99cafc152244af753f735de768cd75f"],
        ["data_batch_2", "d4bba439e000b95fd0a9bffe97cbabec"],
        ["data_batch_3", "54ebc095f3ab1f0389bbae665268c751"],
        ["data_batch_4", "634d18415352ddfa80567beed471001a"],
        ["data_batch_5", "482c414d41f54cd18b22e5b47cb7c3cb"],
    ]

    test_list = [
        ["test_batch", "40351d587109b95175f43aff81a1287e"],
    ]
    meta = {
        "filename": "batches.meta",
        "key": "label_names",
        "md5": "5ff9c542aee3614f3951f8cda6e48888",
    }

    def __init__(
        self,
        root: str,
        train: bool = True,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
    ) -> None:

        super().__init__(root, transform=transform, target_transform=target_transform)

        self.train = train  # training set or test set

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it"
            )

        if self.train:
            downloaded_list = self.train_list
        else:
            downloaded_list = self.test_list

        self.data: Any = []
        self.targets = []

        # now load the picked numpy arrays
        for file_name, checksum in downloaded_list:
            file_path = os.path.join(self.root, self.base_folder, file_name)
            with open(file_path, "rb") as f:
                entry = pickle.load(f, encoding="latin1")
                self.data.append(transform(Image.fromarray(entry["data"])))
                if "labels" in entry:
                    self.targets.extend(entry["labels"])
                else:
                    self.targets.extend(entry["fine_labels"])

        self.data = np.vstack(self.data).reshape(-1, 3, 32, 32)

        self._load_meta()

    def _load_meta(self) -> None:
        path = os.path.join(self.root, self.base_folder, self.meta["filename"])
        if not check_integrity(path, self.meta["md5"]):
            raise RuntimeError(
                "Dataset metadata file not found or corrupted."
                + "You can use download=True to download it"
            )
        with open(path, "rb") as infile:
            data = pickle.load(infile, encoding="latin1")
            self.classes = data[self.meta["key"]]
        self.class_to_idx = {_class: i for i, _class in enumerate(self.classes)}

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.data[index], self.targets[index]
        return img, target

    def __len__(self) -> int:
        return len(self.data)

    def _check_integrity(self) -> bool:
        root = self.root
        for fentry in self.train_list + self.test_list:
            filename, md5 = fentry[0], fentry[1]
            fpath = os.path.join(root, self.base_folder, filename)
            if not check_integrity(fpath, md5):
                return False
        return True

    def download(self) -> None:
        if self._check_integrity():
            print("Files already downloaded and verified")
            return
        download_and_extract_archive(
            self.url, self.root, filename=self.filename, md5=self.tgz_md5
        )

    def extra_repr(self) -> str:
        split = "Train" if self.train is True else "Test"
        return f"Split: {split}"



# Copyright (C) 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import logging

import keras

from openfl.federated import KerasTaskRunner
from src.model import CNNModel

logger = logging.getLogger(__name__)

class CNNTaskruner(KerasTaskRunner):
    """A basic convolutional neural network model."""

    def __init__(self, **kwargs):
        """
        Initializes the TaskRunner instance. Builds the Keras model, initializes required tensors
        for all publicly accessible methods that could be called as part of a task
        and initializes the logger.

        Args:
            **kwargs: Arbitrary keyword arguments passed to the superclass and used for
                model building.

        Attributes:
            model (keras.Model): The Keras model built using the provided feature shape
                and number of classes.
            logger (logging.Logger): Logger instance for logging information.

        Methods:
            build_model: Constructs the Keras model.
            initialize_tensorkeys_for_functions: Initializes tensor keys for various functions.
            get_train_data_size: Returns the size of the training dataset.
            get_valid_data_size: Returns the size of the validation dataset.
        """
        super().__init__(**kwargs)

        self.model = self.build_model(self.feature_shape, self.data_loader.num_classes, **kwargs)

        self.initialize_tensorkeys_for_functions()

        self.model.summary(print_fn=logger.info)

    def build_model(self,
                    input_shape,
                    num_classes,
                    conv_kernel_size=(4, 4),
                    conv_strides=(2, 2),
                    conv1_channels_out=16,
                    conv2_channels_out=32,
                    final_dense_inputsize=100,
                    **kwargs):
        """
        Builds and compiles a Convolutional Neural Network (CNN) model.

        Args:
            input_shape (tuple): Shape of the input data (height, width, channels).
            num_classes (int): Number of output classes.
            conv_kernel_size (tuple, optional): Size of the convolutional kernels.
                Defaults to (4, 4).
            conv_strides (tuple, optional): Strides of the convolutional layers. Defaults to (2, 2).
            conv1_channels_out (int, optional): Number of output channels for the first
                convolutional layer. Defaults to 16.
            conv2_channels_out (int, optional): Number of output channels for the second
                convolutional layer. Defaults to 32.
            final_dense_inputsize (int, optional): Number of units in the final dense layer
                before the output layer. Defaults to 100.
            **kwargs: Additional keyword arguments.
        Returns:
            keras.Model: Compiled CNN model.
        """
        inputs = keras.Input(shape=input_shape)
        outputs = keras.layers.Conv2D(conv1_channels_out,
                                    kernel_size=conv_kernel_size,
                                    strides=conv_strides,
                                    activation='relu',
                                    input_shape=input_shape)(inputs)
        outputs = keras.layers.Conv2D(conv2_channels_out,
                         kernel_size=conv_kernel_size,
                         strides=conv_strides,
                         activation='relu')(outputs)

        outputs = keras.layers.Flatten()(outputs)

        outputs = keras.layers.Dense(final_dense_inputsize, activation='relu')(outputs)

        outputs = keras.layers.Dense(num_classes, activation='softmax')(outputs)

        model = CNNModel(inputs, outputs)

        model.compile(loss="categorical_crossentropy",
                      optimizer="adam",
                      metrics=["accuracy"])

        return model



# Copyright (C) 2024 Intel Corporation
# Licensed subject to the terms of the separately executed evaluation license agreement between
# Intel Corporation and you.

from openfl.federated import PyTorchDataLoader


class TemplateDataLoader(PyTorchDataLoader):
    """Template dataloader for PyTorch.
    This class should be used as a template to create a custom DataLoader for your specific dataset.
    After generating this template, you should:
    1. Implement the `load_dataset` function to load your data.
    2. Modify the `plan.yaml` file to use this DataLoader.
    The `plan.yaml` modifications should be done under the `<workspace>/plan/plan.yaml` section:
    ```
    data_loader:
        defaults: plan/defaults/data_loader.yaml
        template: src.dataloader.TemplateDataLoader # Modify this line appropriately if you
                  change the class name
        settings:
            # Add additional arguments (such as batch_size) that you wish
            # to pass through `def __init__():`
            # You do not need to pass in data_path here. It will be set by the collaborators
    ```
    `batch_size` is passed to the `super().`__init__` method to ensure that the superclass
    is properly initialized with the specified batch size.
    After calling `super().__init__`, define `self.X_train`, `self.y_train`,
    `self.X_valid`, and `self.y_valid`.

    Note: If data_path is None, the dataloader is being initialized for model creation only,
    and no data will be loaded.
    """

    def __init__(self, data_path=None, batch_size=32, **kwargs):
        """Initialize the data loader.
        Args:
            data_path: The file path to the data at the respective collaborator site.
                       If None, initialize for model creation only.
            batch_size: The batch size of the data loader.
            **kwargs: Additional arguments that may be defined in `plan.yaml`
        """
        super().__init__(batch_size, **kwargs)

        # Define required attributes for minimal initialization
        # These should be replaced with appropriate values for your specific dataset
        self.feature_shape = [1, 28, 28]  # Example shape [channels, height, width] for PyTorch
        self.num_classes = 10  # Example number of classes

        # If data_path is None, this is being used for model initialization only
        if data_path is None:
            return

        # Load the dataset using the provided data_path and any additional kwargs.
        X_train, y_train, X_valid, y_valid = load_dataset(data_path, **kwargs)

        # Assign the loaded data to instance variables.
        self.X_train = X_train
        self.y_train = y_train
        self.X_valid = X_valid
        self.y_valid = y_valid

    def get_feature_shape(self):
        """Returns the shape of an example feature array.

        Returns:
            list: The shape of an example feature array.
        """
        return self.feature_shape

    def get_num_classes(self):
        """Returns the number of classes for classification tasks.

        Returns:
            int: The number of classes.
        """
        return self.num_classes


def load_dataset(data_path, **kwargs):
    """
    Load your dataset here.
    This function should be implemented to load the dataset from the given `data_path`.
    You can use additional arguments passed via `**kwargs` if necessary.
    Args:
        data_path (str): Path to the data directory.
        **kwargs: Additional arguments that may be defined in `plan.yaml`
    Returns:
        Tuple containing:
        - numpy.ndarray: The training data.
        - numpy.ndarray: The training labels.
        - numpy.ndarray: The validation data.
        - numpy.ndarray: The validation labels.
    """
    # Implement dataset loading logic here and return the appropriate data.
    # Replace the following placeholders with actual data loading code.
    X_train = None  # Placeholder for training data.
    y_train = None  # Placeholder for training labels.
    X_valid = None  # Placeholder for validation data.
    y_valid = None  # Placeholder for validation labels.

    return X_train, y_train, X_valid, y_valid


raise NotImplementedError("Use <workspace>/src/dataloader.py template to "
                          "create a custom dataloader. Then remove this line.")



import os
import sys
import hashlib
from flwr_datasets import FederatedDataset
from flwr_datasets.partitioner import IidPartitioner
from PIL import Image
import numpy as np
from tqdm import tqdm

EXPECTED_DATASET_HASH = 'ee5342b063299eda95304b0e0388e403a3c65ce5a07cabea4c62a3efdb419a5983c7bb204e55d771ae643642ed440182'

def verify_data_hash(dataset, expected_hash):
    """Verify the hash of the entire dataset."""
    calculated_hash = hash_dataset(dataset)
    if calculated_hash != expected_hash:
        raise ValueError(f'Hash mismatch: {calculated_hash} != {expected_hash}')
    print("Dataset hash verification successful.")

def hash_dataset(dataset):
    """Hash the contents of a Dataset."""
    hash_obj = hashlib.sha384()
    for example in dataset:
        img_array = np.array(example['img'])
        label = example['label']
        # Convert image array and label to bytes
        img_bytes = img_array.tobytes()
        label_bytes = bytes([label])
        # Update hash with image and label bytes
        hash_obj.update(img_bytes)
        hash_obj.update(label_bytes)
    return hash_obj.hexdigest()

def load_and_verify_dataset(dataset_name, expected_hash):
    """Load the entire dataset and verify its hash."""
    # Initialize FederatedDataset with a dummy partitioner
    dummy_partitioner = IidPartitioner(num_partitions=1)  # Dummy partitioner
    fds = FederatedDataset(
        dataset=dataset_name,
        partitioners={"train": dummy_partitioner},
    )

    # Load the entire dataset
    full_dataset = fds.load_partition(0)  # Load the entire dataset as a single partition

    # Hash the entire dataset
    verify_data_hash(full_dataset, expected_hash)


def main(num_partitions):
    # Directory to save the partitions
    save_dir = 'data'

    # Ensure the save directory exists
    os.makedirs(save_dir, exist_ok=True)

    # Load and verify the entire dataset
    load_and_verify_dataset("uoft-cs/cifar10", EXPECTED_DATASET_HASH)

    # Initialize FederatedDataset with the actual partitioner
    partitioner = IidPartitioner(num_partitions=num_partitions)
    fds = FederatedDataset(
        dataset="uoft-cs/cifar10",
        partitioners={"train": partitioner},
    )

    # Download, split, and save the dataset
    for partition_id in range(num_partitions):
        partition = fds.load_partition(partition_id)
        partition_train_test = partition.train_test_split(test_size=0.2, seed=42)

        # Save partition data
        partition_dir = os.path.join(save_dir, f"{partition_id+1}")
        os.makedirs(partition_dir, exist_ok=True)

        for split, dataset in partition_train_test.items():
            split_dir = os.path.join(partition_dir, split)
            os.makedirs(split_dir, exist_ok=True)

            for idx, example in enumerate(tqdm(dataset, desc=f"Processing partition {partition_id+1}, {split}")):
                img_array = np.array(example['img'])
                label = example['label']
                label_dir = os.path.join(split_dir, str(label))
                os.makedirs(label_dir, exist_ok=True)

                # Save the image
                img = Image.fromarray(img_array)
                img_path = os.path.join(label_dir, f"{idx}.png")
                img.save(img_path)

    print("Dataset downloaded, verified, split, and saved successfully.")

if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: python setup_data.py <num_partitions>")
        sys.exit(1)

    num_partitions = int(sys.argv[1])
    main(num_partitions)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from openfl.interface.aggregation_functions.adagrad_adaptive_aggregation import (
    AdagradAdaptiveAggregation,
)
from openfl.interface.aggregation_functions.adam_adaptive_aggregation import AdamAdaptiveAggregation
from openfl.interface.aggregation_functions.core import AggregationFunction
from openfl.interface.aggregation_functions.fed_bagging import FedBaggingXGBoost
from openfl.interface.aggregation_functions.fedcurv_weighted_average import FedCurvWeightedAverage
from openfl.interface.aggregation_functions.geometric_median import GeometricMedian
from openfl.interface.aggregation_functions.median import Median
from openfl.interface.aggregation_functions.secure_weighted_average import SecureWeightedAverage
from openfl.interface.aggregation_functions.weighted_average import WeightedAverage
from openfl.interface.aggregation_functions.yogi_adaptive_aggregation import YogiAdaptiveAggregation



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from openfl.component.aggregator.aggregator import Aggregator
from openfl.component.aggregator.straggler_handling import (
    CutoffTimePolicy,
    PercentagePolicy,
    StragglerPolicy,
    WaitForAllPolicy,
)



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import importlib
import logging

from flwr.proto import grpcadapter_pb2
from google.protobuf.message import DecodeError

from openfl.protocols import aggregator_pb2

logger = logging.getLogger(__name__)


def flower_to_openfl_message(flower_message, header=None, end_experiment=False):
    """
    Convert a Flower MessageContainer to an OpenFL InteropMessage.

    This function takes a Flower MessageContainer and converts it into an OpenFL InteropMessage.
    If the input is already an OpenFL InteropMessage, it returns the input as-is.

    Args:
        flower_message (grpcadapter_pb2.MessageContainer or aggregator_pb2.InteropMessage):
            The Flower message to be converted. It can either be a Flower MessageContainer
            or an OpenFL InteropMessage.
        header (aggregator_pb2.MessageHeader, optional):
            An optional header to be included in the OpenFL InteropMessage. If provided,
            it will be copied to the InteropMessage's header field.

    Returns:
        aggregator_pb2.InteropMessage: The converted OpenFL InteropMessage message.
    """
    if isinstance(flower_message, aggregator_pb2.InteropMessage):
        # If the input is already an OpenFL message, return it as-is
        return flower_message
    else:
        # Check if the Flower message can be deserialized, log a warning if not
        try:
            deserialized_message = deserialize_flower_message(flower_message)
            if deserialized_message is None:
                logger.warning("Failed to introspect Flower message.")
        except Exception as e:
            logger.warning(f"Exception during Flower message introspection: {e}")

        # Create the OpenFL message
        openfl_message = aggregator_pb2.InteropMessage()
        # Set the MessageHeader fields based on the provided sender and receiver
        if header:
            openfl_message.header.CopyFrom(header)

        # Serialize the Flower message and set it in the OpenFL message
        serialized_flower_message = flower_message.SerializeToString()
        openfl_message.message.npbytes = serialized_flower_message
        openfl_message.message.size = len(serialized_flower_message)

        # Add flag to check if experiment has ended
        openfl_message.metadata.update({"end_experiment": str(end_experiment)})
        return openfl_message


def openfl_to_flower_message(openfl_message):
    """
    Convert an OpenFL InteropMessage to a Flower MessageContainer.

    This function takes an OpenFL InteropMessage and converts it into a Flower MessageContainer.
    If the input is already a Flower MessageContainer, it returns the input as-is.

    Args:
        openfl_message (aggregator_pb2.InteropMessage or grpcadapter_pb2.MessageContainer):
            The OpenFL message to be converted. It can either be an OpenFL InteropMessage
            or a Flower MessageContainer.

    Returns:
        grpcadapter_pb2.MessageContainer: The converted Flower MessageContainer.
    """
    if isinstance(openfl_message, grpcadapter_pb2.MessageContainer):
        # If the input is already a Flower message, return it as-is
        return openfl_message
    else:
        # Deserialize the Flower message from the DataStream npbytes field
        flower_message = grpcadapter_pb2.MessageContainer()
        flower_message.ParseFromString(openfl_message.message.npbytes)
        return flower_message


def deserialize_flower_message(flower_message):
    """
    Deserialize the grpc_message_content of a Flower message using the module and class name
    specified in the metadata.

    Args:
        flower_message: The Flower message containing the metadata and binary content.

    Returns:
        The deserialized message object, or None if deserialization fails.
    """
    # Access metadata directly
    metadata = flower_message.metadata
    module_name = metadata.get("grpc-message-module")
    qualname = metadata.get("grpc-message-qualname")

    # Import the module
    try:
        module = importlib.import_module(module_name)
    except ImportError as e:
        print(f"Failed to import module: {module_name}. Error: {e}")
        return None

    # Get the message class
    try:
        message_class = getattr(module, qualname)
    except AttributeError as e:
        print(f"Failed to get message class '{qualname}' from module '{module_name}'. Error: {e}")
        return None

    # Deserialize the content
    try:
        message = message_class.FromString(flower_message.grpc_message_content)
    except DecodeError as e:
        print(f"Failed to deserialize message content. Error: {e}")
        return None

    return message



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Common functions for gRPC transport."""

import logging
import threading
from concurrent.futures import ThreadPoolExecutor

import grpc

from openfl.protocols import aggregator_pb2

logger = logging.getLogger(__name__)


MAX_METADATA_SIZE_BYTES = 32 * 2**20
MAX_MESSAGE_LENGTH_BYTES = 2**30

DEFAULT_CHANNEL_OPTIONS = [
    ("grpc.max_metadata_size", MAX_METADATA_SIZE_BYTES),
    ("grpc.max_send_message_length", MAX_MESSAGE_LENGTH_BYTES),
    ("grpc.max_receive_message_length", MAX_MESSAGE_LENGTH_BYTES),
]


def synchronized(func):
    """Executes `func` synchronously in a threading lock."""
    _lock = threading.Lock()

    def wrapper(self, *args, **kwargs):
        with _lock:
            return func(self, *args, **kwargs)

    return wrapper


def create_insecure_channel(uri) -> grpc.Channel:
    """Creates an insecure gRPC channel."""
    return grpc.insecure_channel(uri, options=DEFAULT_CHANNEL_OPTIONS)


def create_tls_channel(
    uri, root_certificate, require_client_auth, certificate, private_key
) -> grpc.Channel:
    """
    Creates a TLS-based gRPC channel.

    Args:
        uri (str): The uniform resource identifier for the secure channel.
        root_certificate (str): The Certificate Authority filename.
        require_client_auth (bool): True enables client-side
            authentication, i.e. mTLS.
        certificate (str): The client certificate filename from the
            collaborator (signed by the certificate authority).
        private_key (str): The private key filename for the client
            certificate.

    Returns:
        grpc.Channel: A secure gRPC channel object
    """
    with open(root_certificate, "rb") as f:
        root_certificate_b = f.read()

    if not require_client_auth:
        logger.warning("Client-side authentication is disabled.")
        private_key_b = None
        certificate_b = None
    else:
        with open(private_key, "rb") as f:
            private_key_b = f.read()
        with open(certificate, "rb") as f:
            certificate_b = f.read()

    credentials = grpc.ssl_channel_credentials(
        root_certificates=root_certificate_b,
        private_key=private_key_b,
        certificate_chain=certificate_b,
    )

    return grpc.secure_channel(uri, credentials, options=DEFAULT_CHANNEL_OPTIONS)


def create_grpc_server(
    uri: str,
    use_tls: bool,
    private_key: str,
    certificate: str,
    root_certificate: str,
    require_client_auth: bool,
    root_certificate_refresher_cb=None,
) -> grpc.server:
    """Creates a gRPC server.

    Args:
        uri (str): Address for the server (including port number).
        use_tls (bool): If set, the server will use TLS.
        private_key (str): Path to the private key file, required when TLS is enabled.
        certificate (str): Path to the certificate file, required when TLS is enabled.
        root_certificate (str): Path to the root certificate file, required when TLS is enabled.
        require_client_auth (bool): If set, client-side authentication is mandated.
            Only valid when TLS is enabled.
        root_certificate_refresher_cb (function): Callback function to refresh the root certificate.

    Returns:
        grpc.server instance.
    """
    server = grpc.server(ThreadPoolExecutor(max_workers=8), options=DEFAULT_CHANNEL_OPTIONS)

    if not use_tls:
        logger.warning("gRPC is running on insecure channel with TLS disabled.")
        port = server.add_insecure_port(uri)
        logger.info("Insecure port: %s", port)

    else:
        with open(private_key, "rb") as f:
            private_key_b = f.read()
        with open(certificate, "rb") as f:
            certificate_b = f.read()
        with open(root_certificate, "rb") as f:
            root_certificate_b = f.read()

        cert_config = grpc.ssl_server_certificate_configuration(
            ((private_key_b, certificate_b),), root_certificates=root_certificate_b
        )

        def certificate_configuration_fetcher():
            root_cert = root_certificate_b
            if root_certificate_refresher_cb:
                root_cert = root_certificate_refresher_cb()
            return grpc.ssl_server_certificate_configuration(
                ((private_key_b, certificate_b),), root_certificates=root_cert
            )

        if not require_client_auth:
            logger.warning("Client-side authentication is disabled.")
        server_credentials = grpc.dynamic_ssl_server_credentials(
            cert_config,
            certificate_configuration_fetcher,
            require_client_authentication=require_client_auth,
        )
        server.add_secure_port(uri, server_credentials)

    return server


def create_header(sender, receiver, federation_uuid, single_col_cert_common_name):
    """Creates a header for gRPC messages.

    Args:
        sender (str): The sender of the message.
        receiver (str): The receiver of the message.
        federation_uuid (str): The UUID of the federation.
        single_col_cert_common_name (str): The common name on the collaborator's certificate.

    Returns:
        aggregator_pb2.MessageHeader: The header for gRPC messages.
    """
    return aggregator_pb2.MessageHeader(
        sender=sender,
        receiver=receiver,
        federation_uuid=federation_uuid,
        single_col_cert_common_name=single_col_cert_common_name,
    )
